package contextG

import static org.junit.Assert.*

class Base {
    def a() {
       "base a" 
    }
    
    static b() {
        "base b"
    }
    
    def c() {
        "base c" 
    }
    
    static {
        layer(Layer.BaseLayer) {
            a = {
                "base layer a"
            }
            b = {
                "base layer b"
            }
            c = {
                "base layer c"
            }
        }
    }
}

class Derived extends Base {
    def a() {
        "derived a" 
    }
    
    static b() {
        "derived b"
    }
     
    static {
        layer(Layer.DerivedLayer) {
            a = {
                "derived layer a"
            }
            b = {
                "derived layer b"
            }
        }
    }
}

// Demonstrates how layer work with inheritance
// A layer can be interpreted as half step inheritance:
// it overrides all methods of the base classes but leaves method unmodified, 
// if they are redefined in derived classes
class Inheritance {
	static void main(args) {
		def b = new Base(), d = new Derived()

		// a layer overrides the instance methods of the same class
		assertEquals "base a", b.a()
		with(Layer.BaseLayer) {
			assertEquals "base layer a", b.a()
			assertEquals "base layer c", b.c()
		}
		
		// a layer does NOT override instance methods of derived classes if they are redefined there
		// this ensures that derived classes can always rely on their inner assumptions
		assertEquals "derived a", d.a()
		assertEquals "base c", d.c()
		with(Layer.BaseLayer) {
			// a is overriden in derived class
			assertEquals "derived a", d.a()
			// but c is not overriden and therefore layered
			assertEquals "base layer c", d.c()
		}
		// but of course methods can be redefined in their own layer
		with(Layer.DerivedLayer) {
			assertEquals "derived layer a", d.a()
			assertEquals "base c", d.c()
		}
		
		// same with static methods
		assertEquals "base b", Base.b()
		with(Layer.BaseLayer) {
			assertEquals "base layer b", Base.b()
		}
		
		// they are only redefined if the layers are defined after the normally executed method has been defined
		assertEquals "derived b", Derived.b()
		with(Layer.BaseLayer) {
			assertEquals "derived b", Derived.b()
		}
		with(Layer.DerivedLayer) {
			assertEquals "derived layer b", Derived.b()
		}
	}
}